/* 
 * File:   SessionManager.h
 * Author: Miha Nedok <mike@mike.si>
 *
 * Created on April 8, 2013, 3:14 AM
 */


#include "SkynetLib.h"
#include <zdb.h>

#ifndef SESSIONMANAGER_H
#define	SESSIONMANAGER_H


#define SESSION_SERIALIZEDFORMAT_VERSION	1000

//// Application's Pool
//typedef struct {
//  GPtrArray* connections;
//  GMutex lock;
//  time_t lastUsed;
//  long useCount;
//}mongoConnPoolEntry;


typedef struct {
  GMutex lock;
}userTransactionLock;

//// Connection Pool DB connectionEntry
//typedef struct {
//  bool inUse;
//  short state;
//  size_t useCount;
//  DBClientConnection *connection; // the real connection
//  mongoConnPoolEntry *pool;
//  int connId;
//}mongoConnection;

//typedef mongoConnection* ptrMongoConnectionEntry;

typedef struct _keyStore {
  long lastUsed;
  char *passwordHash;
}keyStore;

// Locks for our structures
typedef struct _sessionLocks {
  GMutex tableLock;
  GMutex connectionsLock;
}sessionLocks;

#ifdef __cplusplus
#include <cstdlib>
#include <iostream>
#include <string>
#include <vector>
#include <exception>
#include <ostream>
#include <sstream>
#include <thread>
#include <vector>


#include "MongoIncludes.h"

using namespace std;


class SkynetJsonBuilder {
public:
  SkynetJsonBuilder(){
    _r = json_object();
    // TODO ... Exception for allocation failure.
    if(!_r){
      
    }
  }
  ~SkynetJsonBuilder(){
    
  }
  
  SkynetJsonBuilder& append(const char *fieldName, const char *val){
    if (!val || !strlen(val)){
      json_object_set_new(_r, fieldName,json_null());
    } else {
      json_object_set_new(_r, fieldName,json_string(val));
    }
    return *this;
  }

  SkynetJsonBuilder& appendNull(const char *fieldName){
    json_object_set_new(_r, fieldName,json_null());
    return *this;
  }

  SkynetJsonBuilder& append(const char *fieldName,json_t* j){
    json_object_set_new(_r, fieldName,j);
    return *this;
  }
  
  SkynetJsonBuilder& append(const char *fieldName,string *str){
    json_object_set_new(_r, fieldName,json_string(str->c_str()));
    return *this;
  }
  
  SkynetJsonBuilder& append(const char *fieldName,int n){
    json_object_set_new(_r, fieldName,json_integer(n));
    return *this;
  }
  
  SkynetJsonBuilder& append(const char *fieldName,bool b){
    json_object_set_new(_r, fieldName,json_boolean(b));
    return *this;
  }
  
  SkynetJsonBuilder& append(const char *fieldName,double d){
    json_object_set_new(_r, fieldName,json_real(d));
    return *this;
  }
  
  SkynetJsonBuilder& append(json_t *src_json){
    json_object_update(_r, src_json);
    return *this;
  }
  
  SkynetJsonBuilder& appendBool(const char *fieldName,bool bval){
    json_object_set_new(_r,fieldName,json_boolean(bval));
    return *this;
  }
  
  // Return the json_t ptr
  json_t* getJson(){
    return this->_r;
  }
  
  char* jsonBuf(){
    return json_dumps(_r,JSON_COMPACT);
  }
  
  // Destroy the Json object
  void destroy(){
    json_decref(this->_r);
  }
  
  std::string jsonString(){
    std::string(json_dumps(_r,JSON_COMPACT));
  }
private:
  /**
   * The Resulting object
   */
  json_t *_r;
};



/// Custom Exception Class with auto logging.
class SkynetException : public std::exception {
public:
  SkynetException(const char *errMessage,int errCode){
    errMessage_ = errMessage;
    errCode_    = errCode;
    syslog(LOG_ERR,"SkynetException: [%s] Code: [%d]",errMessage_,errCode_);
  }
  SkynetException(const char *errMessage,int errCode,int logLevel){
    errMessage_ = errMessage;
    errCode_    = errCode;
    syslog(logLevel,"SkynetException: [%s] Code: [%d]",errMessage_,errCode_);
  }
  ~SkynetException(){
    
  }
  const char* what() const throw() {
    return errMessage_;
  }
  const int getErrorCode() const throw(){
    return errCode_;
  }
private:
  const char *errMessage_;
  int errCode_;
};


class SessionManager {
public:
  SessionManager();
  SessionManager(const SessionManager& orig);
  virtual ~SessionManager();
  
  /**
   * Does session exist ?
   */
  virtual bool sessionExists(const char* sessionId);
  /**
   * Extend a given session for given amount in seconds
   */
  virtual void extendSession(const char* sessionId,size_t extendInSeconds);
  /**
   * Extend give session.
   * Performs all necesaary locks
   */
  virtual void extendSessionIfExists(const char *sessionId);
  
  /**
   * Creates a new session. With given sessionId, this is the most common scenario as opposed that
   * we create the sessionId here on this level.
   *
   *
   *  @param userid
   *  @param sessionId
   *  @param data
   *  @param remoteAddr
   *  @param additional
   *  @param permanent
   *  @param newSession
   *
   *  @return success
   */
  virtual bool createSession(const char* userid,char* sessionId,const char *data,const char* remoteAddr,void* additional,int permanent,SessionEntry** newSession);
  /**
   *
   *
   */
  virtual bool readSession(const char *sessionId,char **sessionData, size_t *sessionDataLen);
  
  /**
   *
   *
   */
  virtual bool writeSession(const char *sessionId,const char *sessionData,const char *remoteAddr);
  
  /**
   *
   *
   *  @param sessionid
   *
   *  @return success
   */
  virtual bool destroySession(const char *sessionid);
  /**
   *  Logout all user sessions
   *
   *  @param userId
   *  @param appId
   *  @param remoteAddr
   *  @param adminId    admin which logouts.
   *  @param reason     reason for logout.
   */
  virtual void logoutUserSessions(const char *userId,const char *appId,const char *remoteAddr,const char *adminId,const char *reason);
  /**
   *  The user has changed - refresh all data
   *
   *  @param userId
   */
  virtual void userChanged(const char *userId);
  /**
   *
   *
   *  @param sessionId
   */
  virtual void sessionSetAdmin(const char *sessionId);
  /**
   *
   *
   *  @param sessionId
   */
  virtual void sessionUnsetAdmin(const char *sessionId);
  

  /**
   *  Clean expired and invalid Sessions.
   */
  virtual void cleanerWalker();
  

  virtual void subscribeChannel(const char *channelId);
  virtual void unsubscribeChannel(const char *channelId);

  
  // Try to DB auth user with password.
  // Useful for commands or actions which require auth to be present
  // and verified inside valid session.
  // virtual bool dbAuthUser(const char *dbSetting,const char *username, const char *password, const char *deviceToken, const char *remoteAddr,const char *deviceId,long long shortUsername);
  
  virtual bool isAnonUser(const char *sessionId);
  virtual bool isAdmin(const char *sessionId);
  virtual bool isPermanent(const char *sessionId);
  
  // Time Authoziation
  virtual bool isTimeAuthorized(const char *sessionId); // Checks if or expired.
  virtual bool isTimeAuthorizedExpired(const char *sessionId);
  virtual bool requestTimeAuthorization(const char *sessionId,const char *appId,char **authorizationToken);
  virtual int  verifyTimeAuthorization(const char *sessionId,const char *userId,const char *password,
                                       const char *authorizationToken,const char *remoteAddr,const char *clientInfo,const char *appId);
  
  // End Time Authorization
  
  virtual GList* getSessionAdmins();
  
  /**
   *
   */
  virtual bool updateSession(const char *sessionId,const char *userId,const char *remoteAddr);
  
  /**
   * Push a session to tables, the passed in values are not copied except data!
   * Primarily used by unserializer.
   * For data never push a copy but rather the source pointer, because data is being copied
   * to GString structure.
   */
  virtual void pushSession(char* userid,char *sessionId,char *data,char *remoteAddr,int cTime, int laTime,int permanent,gboolean isAdmin, const char *userType,
                           char *companyId,char *companyType);
  
  /**
   * Create a new session and we create a new sessionId
   */
  //virtual bool createSession(const char* userid,const char* remoteAddr,void* additional,SessionEntry** newSession);
  /**
   * Delete and forget a session
   */
  virtual void deleteSession(const char* sessionId);
  
  
  /**
   Deletes the session inside the manager.
   Should be called only by serializer.

   @param sessionId The Id.
   */
  virtual void deleteSessionAuto(const char *sessionId);
  
  /**
   *  Return number of current sessions.
   *
   *  @return number of session.
   */
  virtual size_t numSessions();
  
  virtual void printVersion();
  
  /**
   *  Serializer for session data.
   */
  virtual void serializeSessionData();
  /**
   *
   */
  virtual void unserializeSessionData();
  
  
  
  /**
   * Add a new connection to table, does not check if
   * currently a connection exists
   */
  virtual void addConnection(int connectionId,gpointer connection,const char* sessionId);
  /**
   *
   *
   *  @param userId
   *
   *  @return
   */
  virtual GHashTable *getConnectionsByUserId(const char *userId);
  /**
   *
   *
   *  @param userId
   *
   *  @return
   */
  virtual GPtrArray *getConnectionsByUserIdExclusive(const char *userId);
  
  /**
   * Delete a connection from our map of current connections
   */
  virtual void deleteConnection(int connectionId);
  /**
   *
   *
   *  @param connectionId
   *
   *  @return
   */
  virtual bool haveConnectionById(int connectionId);
  /*
   * Is an user offline
   */
  virtual bool userIsOffline(const char *userId);
  
  /**
   * Fetch the userId for given sessionId
   */
  virtual const char* userIdBySessionId(const char* sessionId);
  /**
   *
   */
  virtual const char* getSessionIdByConnectionId(long connectionId);
  /**
   *
   *
   *  @param userId
   *
   *  @return
   */
  GPtrArray *getSessionsByUserId(const char* userId);
  /**
   *
   *
   *  @param sessionId
   *
   *  @return
   */
  virtual SessionEntry* getSessionById(const char *sessionId);
  
  /**
   *
   *
   *  @param originatorId
   *  @param recipientId
   *  @param resourceId
   *  @param recipientList
   *  @param resourceType
   *  @param priority
   *  @param additionalData
   */
  virtual void postNotification(const char *originatorId,const char *recipientId,const char *resourceId,GPtrArray *recipientList,int resourceType,int priority,void *additionalData);
  
  //virtual void session_manager_ws_push_for_user(const char *userId,const char *payload);
  
  /**
   * Network stuff
   */
  /**
   * Send a message to a single connected user
   */
  virtual void pushMessageForUser(const char* userId,const char* message, size_t len);
  /**
   * Send a message to multiple users.
   */
  void pushMessageForMultipleUsers(GPtrArray userList,const char* message,size_t len);
  
  /**
   *
   */
  GHashTable* sessionTable;
  /**
   *
   */
  sessionLocks *locks;
  GHashTable* userToSessionTable;
  GHashTable* connectionMap;
  GHashTable* connectionSessionMap; // ID of connection to SessionId
  GHashTable* secTable; // Security table - for private keys storage
  GHashTable *adminTable; // Table holding admin sessions
  GHashTable *chanSubsTable;
  
private:
  
  
  
  
  
  
};

// Default SessionManager Implementation is Server (master node)
class SessionManagerServer: public SessionManager{
public:
  SessionManagerServer();
};
// End Server

// Client Implementation for Clustering

class SessionManagerClient: public SessionManager{
public:
  // Client Specific:
  SessionManagerClient();
  bool zmqConnectServer(const char *host,int port);
  bool zmqSendCommand(void *bbuilder);
  
  // Overload:
  
  bool sessionExists(const char* sessionId);
};
// End Client



// Manager macros

#define GM_LOAD_USER(msg,userId) globalManager->loadUser(msg->appId,userId)
#define GM_GETMONGOCONNECTION(appId) DBClientConnection *connection = new DBClientConnection();connection->connect(globalManager->applicationMongoHost(appId))
// GlobalManager internal macro
// #define _GM_GETMONGOCONNECTION(appId) ScopedDbConnection connection(this->applicationMongoHost(appId))
#define _GM_GETMONGOCONNECTION(appId) DBClientConnection *connection = new DBClientConnection();connection->connect(this->applicationMongoHost(appId))
#define _GM_GETMONGOCONNECTION_N(appId) DBClientConnection *connection = new DBClientConnection();connection->connect(this->applicationMongoHost(appId))

// Global user macros

// Global Manager Object
// Is a global allround object
// for all purposes:
// - user handling
// - common data handling
// - websocket pusher
// - notifications wrapper
namespace skynet {
  
  
  
  typedef struct {
    int rval;
    char *transId;
    GString *errorMessageExtra;
    json_t *extraErrorData;
  }transactionResult;
  
  
  // Single application entry in map
  class AppEntry{
		public:
    char *mongoHost;
    char *mongoDb;
    char *mongoUser;
    char *mongoPass;
    mongocxx::uri mongoUri;
    mongocxx::pool *mongoPool; // Application specific pool
    char *mysqlDb;
    char *mysqlHost;
    char *mysqlUser;
    char *baseUrl;
    char *nchanPublisherUrl;
    ConnectionPool_T dbPool;
  };
  
  
  typedef struct{
    char *apiKey;
    char *apiSecret;
    char *serviceUrl;
  }_smsServiceEntry;
  
  
  // Tasks - Platform shared functionality
  class Tasks {
  public:
    // virtual BSONObj loadTask(SkynetServerMessage *msg,const char *taskId);
    //     virtual bool isMyTask(SkynetServerMessage *msg,BSONObj bTask);
  };
  
  
  class GlobalActivity;
  class Transction;
  class MemoryRegister;
  class MongoConnectionPool;
  
  class GlobalManager {
		private:
    // No need for protection of this resource - is only used as read
    //		std::unordered_map<std::string,AppEntry*> appMap;
    GHashTable *hAppMap; // Faster than unordered_map
    ServiceConfig *mainConfig;
    
    // Holder for UserId to Mutex
    GHashTable *hUserTransactionMutexes;
    
    // Configured SMS services.
    GHashTable *smsServices;
    
    AppEntry *initNewApp(json_t *appConfig);
    void loadApps(const char *appMapFile);
    void loadSmsServices();
    /**
     *  Load a given configuration from database and return ptr to
     *  parsed json structure, if fails then the ptr to json is NULL.
     *
     *  @param settingName Name of the settings in the database.
     *  @param lastUpdateTime The time at which it was last updated.
     *
     *  @return json parsed ptr or NULL if failed
     */
    json_t* loadSettingFromDb(const char *settingName,int lastUpdateTime);
    
    /**
     *  Load Cluster Config from Mongo sys.syssettings and
     *  parse everything into MemoryRegister
     */
    void loadClusterConfig();
    
    inline AppEntry* _fetch_map_app(const char *appId){
      return (AppEntry*)g_hash_table_lookup(this->hAppMap,(gpointer)appId);
    }
    
    inline userTransactionLock* _fetch_user_transaction_mutex(const char *userId){
      return (userTransactionLock*)g_hash_table_lookup(this->hUserTransactionMutexes,(gconstpointer)userId);
    }
    
    int clusterConfigLastUpdate;
    
		public:
    /**
     *
     */
    Tasks *taskManager;
    /**
     *
     */
    GlobalActivity *systemActivity;
    /**
     *
     */
    SessionManager *sessionManager;
    /**
     *  Memory Register Class
     */
    MemoryRegister *memRegister;
    /**
     *  Skynet HTTP Server
     */
    struct MHD_Daemon *httpServer;
    
    /**
     *  Parsed Cluster Config
     */
    json_t *j_clusterConfig;
    
    
    GlobalManager();
    
    // SessionManager should be App provided.
    GlobalManager(SessionManager *sManager);
    GlobalManager(SessionManager *sManager,ServiceConfig *sConfig);
    
    
    // Transaction related
    virtual GMutex* fetchUserTransactionMutex(const char *userId);
    virtual void releaseUserTransactionMutex(GMutex *userTransMutex);
    
    // System Stuff
    Connection_T applicationSql(const char *appId);
    const char* applicationNchanPublisherUrl(const char *appId);
    virtual mongocxx::pool* applicationMongoPool(const char *appId);
    virtual AppEntry* applicationEntry(const char *appId);
    const char* applicationBaseUrl(const char *appId);
    void freeStrVector(vector<string*>*fVec);
    
    // SMS Service
    bool getSmsServiceData(const char *appId,const char *service,char **apiKey,char **apiSecret,char **serviceUrl);
    
    // User Stuff
    bool insertNewUser(SkynetServerMessage *msg);
    bool updateUser(SkynetServerMessage *msg);
    bool activateUser(SkynetServerMessage *msg);
    bool deactivateUser(SkynetServerMessage *msg);
    bool unlockUser(SkynetServerMessage *msg);
    
  //  bool userStructureCheck(BSONObj oUser,const char *appId);
    
    bool passwordVerify(const char *userId,const char *passwordToVerify,const char *appId);
    bool passwordVerify(bsoncxx::document::value vUser,const char *passwordToVerify);
    
    bsoncxx::document::value loadUser(const char *userId,const char *appId);
    
    // Loads Account - no account exists a new one is created with
    // initial balance of 0
    // BSONObj loadUserAccount(const char *ownerId, const char *appId,BSONObj oUser);
    
    
    const char* userCompanyId(const char *userId,const char *appId);
//    static inline const char* userCompanyId(BSONObj oUser){
//      return oUser.getStringField("companyId");
//    }
    
    
    // Global admin
    bool isSuperAdmin(const char *appId, const char *userId);
    bool isSuperAdmin(SkynetServerMessage *msg);
    
    
    // Is Company Admin
    bool isAdmin(const char *appId, const char *userId);
    bool isAdmin(const char *appId, const char *userId, const char *companyId);
    bool isAdmin(SkynetServerMessage *msg,const char *userId);
    bool isAdmin(SkynetServerMessage *msg);
    
    virtual bool changePasswordUser(const char *userId,const char *currentPassword,const char *newPassword, const char *newPasswordConfirm,const char *appId,const char *sessionId);
    virtual bool changePasswordForUser(const char *userId, const char* changerUserId,const char *newPassword,const char *appId);
    virtual bool changePasswordReset(const char *userId,string *newPassword,const char *appId);
    
    // Permissions
    bool permissionMatch(const char *appId, const char *userId,char *perm,...);
    
    // Company Stuff
//    BSONObj loadCompany(const char *companyId, const char *appId);
    
  };
  
  
  typedef struct registryTableEntry{
    GHashTable *table;
    GMutex lock;
  }registryTableEntry;
  typedef registryTableEntry* registryTableEntryP;
  
  /// Thread Safe Memory Register is used to store
  /// pointers stored by integer keys.
  /// It's only for live data.
  /// On shutdown all data gets lost.
  class MemoryRegister {
  public:
    MemoryRegister();
    /**
     *  Try to find data.
     *
     *  @param type Type of data.
     *  @param key  Key for data.
     *
     *  @return data or NULL
     */
    virtual gpointer lookup(int type,const char *key);
    
    /**
     *  Get the whole Table for a Type.
     *
     *  @param type Type of table.
     *
     *  @return the table entry or NULL if not found.
     */
    virtual registryTableEntryP table_for_type(int type);
    
    /**
     *  Run a foreach with given callback.
     *  For the time of the forach the table is automatically locked.
     *
     *  @param type        Type of table
     *  @param foreachFunc Forach function.
     *  @param foreachData Custom data passed for foreach func.
     */
    virtual void foreach_type(int type,gpointer foreachFunc,gpointer foreachData);
    
    /**
     *  Insert new data to register.
     *
     *  @param type Type of data.
     *  @param key  Key for data. The key is always copied.
     *  @param data Data.
     *
     *  @return success
     */
    virtual gboolean insert(int type,const char *key,gpointer data);

    /**
     *  Removes the data from register and returns the ptr to data
     *  so caller can free. The key gets freed automatically.
     *
     *  @param type Type of data
     *  @param key  Key for data.
     *
     *  @return data or NULL if not found.
     */
    virtual gpointer remove(int type,const char *key);
  
  private:
    /**
     *  Fetches a table for type, if not yet existing a 
     *  new str_hash table is created.
     *
     *  @param type Type of data.
     *
     *  @return Table entry or NULL if error occured.
     */
    virtual registryTableEntryP fetch_table_for_type(int type);
    /**
     *  Storage for data maps.
     */
    GHashTable *p_register;
  };
  
  
  enum ActivityId : int {
    USER_LOGIN  = 1,
    USER_PASSWORD_CHANGE = 2,
    USER_LOGOUT = 3,
    USER_PASSWORD_RESET_OK = 3,
    USER_PASSWORD_RESET_ERR = 4
  };
  
  class GlobalActivity {
    
  private:
    GThreadPool *activityPool;
    GlobalManager *globalManager;
    
  public:
    static const int USER_LOGIN													= 1;
    static const int USER_PASSWORD_CHANGE								= 2;
    static const int USER_LOGOUT												= 3;
    static const int USER_PASSWORD_RESET_OK							= 4;
    static const int USER_PASSWORD_RESET_ERR						= 5;
    static const int USER_MICROTRANSACTION_ERR					= 6;
    static const int USER_MICROTRANSACTION_SUCCESS			= 7;
    static const int USER_TRANSACTION_DEBIT_ERR					= 8;
    static const int USER_TRANSACTION_DEBIT_SUCCESS			= 9;
    static const int USER_TRANSACTION_CREDIT_ERR				= 10;
    static const int USER_TRANSACTION_CREDIT_SUCCESS		= 11;
    static const int USER_TRANSACTION_ADDFUNDS_SUCCESS	= 12;
    static const int USER_TRANSACTION_ADDFUNDS_ERR			= 13;
    
    GlobalActivity(GlobalManager *gm);
    ~GlobalActivity();
    void submitActivity(ActivityId activityId,bool success,const char *userId,const char *appId,const char *sessionId,const char *data);
    void submitActivity(int activityId,bool success,const char *userId,const char *appId,const char *sessionId,const char *data);
    void submitActivity(int activityId,bool success,const char *userId,const char *appId,const char *sessionId,const json_t *jdata);
    
  };
  
  /**
   * Mongo Connection Pool Implementation
   * for Skynet and related Projects
   */
//  class MongoConnectionPool {
//  private:
//    GHashTable *connectionsPool;
//    GlobalManager *globalManager;
//  public:
//    MongoConnectionPool(GlobalManager *gm);
//    ~MongoConnectionPool();
//
//    /**
//     * Finds a pool entry or creates a new one. If NULL returned the pool is full
//     * or other problem occured.
//     */
//    ptrMongoConnectionEntry findFreeConnection(const char *appId);
//    void release(ptrMongoConnectionEntry connection);
//    bool initPool(const char *mongoHost, const char *db, const char *un, const char *pw);
//  };
  
};


//void LibSkynetAuthUser(SkynetServerMessage *msg,skynet::GlobalManager *gm,SessionManager *sessionManager);


#endif

#endif	/* SESSIONMANAGER_H */

