//
//  SkynetBank.h
//  LibSkynet
//
//  Created by Miha Nedok
//  Copyright © 2016 Mike. All rights reserved.
//

#ifndef SkynetBank_h
#define SkynetBank_h


#ifndef mpGFreePtr
#define mpGFreePtr(ptr){if(ptr!=NULL){g_free(ptr);}}
#endif

namespace skynet {
  
  enum BankReturnCode : int {
    // Generic OK value
    OK																								= 0,
    // Database Error
    ERR_DB																						= 12,
    // Lock acquire or release error
    ERR_LOCK																					= 13,
    TRANS_ERR_NOT_VALID_PRODUCT												= 1,
    TRANS_ERR_INSUFFICIENT_FUNDS											= 2,
    TRANS_ERR_USER_INVALID														= 3,
    TRANS_ERR_USER_NOTACTIVE													= 4,
    TRANS_ERR_RECIPIENT_INVALID												= 5,
    TRANS_ERR_LOCKING																	= 6,
    TRANS_ERR_INTERNAL_ERR														= 7,
    TRANS_ERR_DATABASE_PROBLEM												= 8,
    TRANS_ERR_ACCOUNT_NOT_OK													= 9,
    TRANS_ERR_INIT_FAILED															= 10,
    // Passed parameters are not correct
    TRANS_ERR_PARAMETERS															= 11,
    // Error with Account
    UPDATE_BALANCE_ERR_ACCT_PROBLEM										= 14,
    ERR_GENERIC																				= 15,
    ERR_ARRAY_EMPTY																		= 16,
    ERR_JSON_DUMP																			= 17
  };
  
  enum TransactionTypeId : int {
    CREDIT_CARD					= 1,
    BANK_TRANSFER				= 2,
    UPN									= 3,
    SMS									= 4,
    PAYPAL							= 5,
    CASH								= 6,
    PREPAID							= 7,
    MICRO								= 8,
    BITCOIN							= 9,
    MONETA							= 10,
    BONUS               = 12
  };
  
  /**
   Declare Transaction Types
   */
  enum TransactionType: int {
    CREDIT = 1,
    DEBIT  = 2
  };
  
  class Amount {
  public:
    Amount(double amount, char *currency);
    ~Amount();
    double amount;
    char *currency;
  };
  
  class OrderItem {
  public:
    char *itemId;
    double amount;
    char *currency;
    int quantity;
    ~OrderItem(){
      if (this->currency){
        g_free(this->currency);
      }
      if (this->itemId){
        g_free(this->itemId);
      }
    }
  };
  
  // Vector for items
  typedef std::vector<OrderItem*> vOrderItems;
  typedef vOrderItems* vOrderItemsPtr;
  
  
  class Order {
  public:
    Order(GlobalManager *gm,const char *appId,const char *currency);
    /**
     *  Destructor
     */
    ~Order(){
      if (this->orderId){
        g_free(this->orderId);
      }
      if (this->orderItems){
        delete this->orderItems;
      }
    }
    /**
     *  Add a single item to order.
     *
     *  @param item Item
     */
    void addItem(OrderItem *item);
    /**
     *  Add a single item to order which is the same currency as order.
     *
     *  @param itemId   Id
     *  @param amount   Amount of Item
     *  @param quantity Quantity
     */
    void addItem(const char *itemId,double amount,int quantity);
    /**
     *  Create a new Order. There have to be items added before with addItem.
     *
     *  @return OrderId
     */
    char* newOrder();
    
    /**
     *  Insert an item into database.
     *
     *  @param item Item to insert.
     *
     *  @return Success.
     */
    gboolean insertOrderItem(OrderItem *item,time_t insertTime);
    
    /**
     * Verify an item object.
     *
     *  @param orderItem
     */
    gboolean verifyOrderItem(OrderItem *item);
    /**
     * Build an array of OrdeItem objects from a json_t Array.
     *
     *  @param arrayOrderItems
     *
     *  @return
     */
    vOrderItemsPtr buildOrderItems(json_t *arrayOrderItems);
    /**
     * Commit the Order into the database
     * and lay foundation for Transaction
     *
     *  @return BankReturnCode
     */
    BankReturnCode commitOrderBegin();
    /**
     *
     *
     *  @param success
     *
     *  @return Return Code.
     */
    BankReturnCode commitOrderEnd(bool success);
    /**
     * Vector of OrderItem objects.
     */
    vOrderItemsPtr orderItems;
    char *orderId;
    double amount; // Sum of all OrderItems
    const char* currency;
    GlobalManager *globalManager;
    const char *appId;
    
    const char *sessionId;
    const char *userId;
    const char *remoteAddr;
    const char *transactionTypeId;
    const char *systemItemId;
  private:
    Connection_T sqlConnection;
    
  };
  
  class Account;
  class UserAccount {
  private:
    const char *appId;
    const char *userId;
    GlobalManager *globalManager;
    double dailyLimitAmount;
    const char *dailyLimitCurrency;
    double weeklyLimitAmount;
    const char* weeklyLimitCurrency;
  public:
    UserAccount(skynet::Account *acct);
    ~UserAccount(){
      if (this->accountId)
        g_free(this->accountId);
    }
    /**
     *  Refresh Account Data from DB
     *
     *  @return BankReturnCode
     */
    BankReturnCode refresh();
    /**
     *
     *  Sets the daily limit of account.
     *
     *  @param amount
     *  @param currency
     *
     *  @return
     */
    bool setDailyLimit(double amount, const char *currency);
    
    virtual bool dailyLimitExceeded(double amount, const char *currency);
    
    virtual double getBalance(const char *currency);
    
    virtual double getBalance(const char *currency,Connection_T con);
    
    virtual BankReturnCode updateBalance(double amount,const char *currency,TransactionType transType,Connection_T con);
    
    // Account's Current Balance
    double balance;
    char* accountId = NULL;
  };
  
  
  // Return object
  class BankMethodResult {
    BankReturnCode rc;
    string* message;
    
  public:
    BankMethodResult(BankReturnCode resultCode,string message);
  };
  
  class MiniBank {
  public:
    
  private:
    const char *mUserId;
    
  };
  
  
  // Account for User
  class Account {
  public:
    UserAccount* account = NULL;
    /*
     End Update Balance return codes
     */
    
    // Constructor
    Account(SessionManager *sm,GlobalManager *gm,const char *userId,const char *appId);
    ~Account();
    
    //		Load account object from database - if not found a new
    //		account is created with empty balance.
    virtual UserAccount* loadAccount(const char *userId, const char *appId);
    
    virtual BankReturnCode refresh();
    
    //
    // Verifies amount for daily limit
    //
    virtual bool dailyLimitExeeded(double amount, const char *currency);
    
    virtual bool isLocked();
    // Check if account has enough funds for given amount
    virtual bool havingFunds(double amount,const char *currencyCode);
    
    virtual double getAccountBalance(const char *currencyCode);
    
    // Load existing account
    // BSONObj loadUserAccountExists(const char *userId, const char *appId);
    
    // Updates the Account Balance after successfull entry into
    // Statements for account.
    virtual BankReturnCode updateBalance(const char *accountId,double amount,
                                         const char *currency, TransactionType transType,bool lockMutex,GMutex *userTransactionMutex);
    
    
    // Push transaction data to Account Statement.
    virtual BankReturnCode addTransactionToStatement(const char *transactionId, double amount,const char *currency,
                                                     TransactionType transType, bool entryFinal);
    
    BankReturnCode stornoTransactionStatement(const char *transactionId);
    
    BankReturnCode fetchStatementsForUser(const char *userId, const char *appId,time_t startTime,time_t endTime,size_t offset);
    
    const char *userId;
    const char *appId;
    const char *accountType;
    bool locked = false;
    GlobalManager *globalManager;
    SessionManager *sessionManager;
  };
  
  class Transaction {
    
  public:
    
    Account* account;
    
    Transaction(SessionManager *sm,GlobalManager *gm,Account *account);
    Transaction(Account *account);
    
    transactionResult* allocResult();
    void							freeResult(transactionResult *result);
    
    int buildTransactionJsonObject();
    
    //
    // Perform basic user check before we allow transaction.
    //
    inline BankReturnCode verifyUserForTransaction(const char *userId, const char *appId){
//      BSONObj oUser = this->globalManager->loadUser(userId,appId);
//      // User not found
//      if (oUser.isEmpty()){
//        return BankReturnCode::TRANS_ERR_USER_INVALID;
//      }
//      // User is not activated
//      if (oUser.getBoolField("active")!=true){
//        return BankReturnCode::TRANS_ERR_USER_NOTACTIVE;
//      }
      return BankReturnCode::OK;
    }
    
    /*
     * Tries to process the given micro transaction.
     * Micro Transactions are from one account to another in the same system.
     * Throws SkynetException
     */
    void microTransaction(const char *sessionId, const char *userId, const char *appId,
                          const char *remoteAddr,const char *itemId, const char *recipientId,
                          double amount, const char *currency,
                          const char *orderId,
                          json_t* jAddtionalData);
    
    /*
     * Server signing of transaction.
     */
    char* signTransaction(const char *userId, const char *appId, const char *transactionId, const char *orderId,
                          double amount, const char *currency,time_t timeOfTransaction);
    //
    // Server signature verification
    //
    bool	verifyTransactionSignature(const char *transactionId);
    
    //virtual initTransaction(const char *userId, const char *appId, const char *password);
    
    //
    //	Try to add given funds to an existing account for user.
    //  Throws SkynetException.
    virtual void addFunds(const char *sessionId, const char *userId, const char *appId,
                          const char *remoteAddr,
                          const char *remoteId, const char *remoteStatus,const char *remoteDescription,
                          TransactionTypeId typeId,
                          json_t *additionalData,
                          const char *orderId,
                          double amount,const char *currency,
                          gchar **transactionId);
    
    // Initialize External Transaction on our endpoint
    virtual transactionResult* externalTransactionInit();
    //
    virtual transactionResult* externalTransactionCommit();
    
    
    // Storno transaction in system
    virtual BankReturnCode stornoTransaction(const char *sessionId, const char *userId, const char *appId,
                                             const char *stornoUserId,
                                             const char *transactionId,int stornoReason,const char *additionalData);
    
    //
    //
    // Throws SkynetException
    virtual void buyCredits(const char *sessionId, const char *userId,const char *remoteAddr,
                            int amountOfCredits,double creditPricing, const char *currency,TransactionTypeId transTypeId,
                            const char *remoteId, const char *remoteStatus, const char *remoteDescription,
                            const char *orderId,
                            json_t *additionalData);
    
    //
    //
    // Throws SkynetException
    virtual void buyWithCredits(const char *sessionId, const char *userId, const char *remoteAddr,const char *appId,
                                const char *buyItemId, const char *beneficiary,const char* amountOfCredits,const char *currency,
                                json_t *additionalData);
    
    
    // Place order in system
    // Throws SkynetException
    virtual void placeOrder(Order *order,const char *sessionId, const char *userId, const char *appId,
                            const char *remoteAddr,json_t *additionalData);
    
    virtual bool eclipseTransactionFund(const char *appId,const char *applicationId,const char *posApiKey, const char *posApiSecret,const char *remoteAddr, const char *mediaId,
                                        const char *remoteId, double amount, const char *currency);
    virtual bool eclipseTransactionPay();
    
  private:
    // Executes given transaction on account - performing all necesary locks
    virtual BankReturnCode _executeTransaction(const char *transactionId,const char *sessionId,
                                               const char *remoteAddr,TransactionTypeId typeId,
                                               const char *beneAccountId,const char *orderAccountId,
                                               const char *referalId, const char *remoteDescription,
                                               const char *remoteId,const char *remoteStatus,
                                               double amount, const char *currency,
                                               const char *orderId,
                                               TransactionType transType,
                                               json_t* jAddtionalData);
    
    SessionManager* sessionManager;
    GlobalManager*  globalManager;
  };
};

#endif /* SkynetBank_h */
