//
//  SkynetHTTP.h
//  LibSkynet
//
//  Created by Miha Nedok on 26/09/16.
//  Copyright © 2016 Mike. All rights reserved.
//
//  SkynetHTTP - HTTP Server Framework
//


#ifndef SkynetHTTP_hpp
#define SkynetHTTP_hpp

#include "SkynetLib.h"

#include <microhttpd.h>

#include <arpa/inet.h>
#include <ifaddrs.h>
#include <net/if.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netdb.h>

/**
 *  Export Everything to Library
 */
BEGIN_EXTERN_C()

/* Generic HTTP Headers */

#define SKYNETHTTP_X_FORWARDED_FOR          "X-Forwarded-For"
#define SKYNETHTTP_USERAGENT                "User-Agent"
#define SKYNETHTTP_COOKIE                   "Cookie"
#define SKYNETHTTP_ORIGIN                   "Origin"
#define SKYNETHTTP_HOST                     "Host"


/* Custom HTTP Header Definitions */

#define SKYNETHTTP_HEADER_NONCE             "Skynet-Nonce"
#define SKYNETHTTP_HEADER_API_SIGNATURE     "Skynet-Api-Signature"
#define SKYNETHTTP_HEADER_CLIENT_ID         "Skynet-Api-Key"
#define SKYNETHTTP_HEADER_SESSION_ID        "Skynet-SessionId"
#define SKYNETHTTP_HEADER_API_VERSION       "Skynet-Api-Version"
#define SKYNETHTTP_HEADER_API_NONCE         "Skynet-Nonce"

/* Nchan Related Headers */

#define SKYNETHTTP_NCHAN_SUBSCRIBER_TYPE    "X-Subscriber-Type"
#define SKYNETHTTP_NCHAN_PUBLISHER_TYPE     "X-Publisher-Type"
#define SKYNETHTTP_NCHAN_PREV_MESSAGE_ID    "X-Prev-Message-Id"
#define SKYNETHTTP_NCHAN_CHANNEL_ID         "X-Channel-Id"
#define SKYNETHTTP_NCHAN_ORIGINAL_URI       "X-Original-URI"


/* API Versions */

#define SKYNETHTTP_JSON_API_VERSION_1       1
#define SKYNETHTTP_JSON_API_VERSION_CURRENT SKYNETHTTP_JSON_API_VERSION_1


/**
 *  Is Called after Request Completes.
 */
typedef void
(*SkynetHTTP_RequestCompleteCallBack) (void *custom_data);

/**
 *
 */
typedef struct skynet_http_response_s{
  json_t *j_response;
  char *rawResponse;
  char *rawHeader;
  GHashTable *headers;
  int responseCode;
  gboolean success;
}skynet_http_response_s;
typedef skynet_http_response_s *skynet_http_response_p;

/**
 *  HTTP Request Message Struct
 */
typedef struct skynet_http_msg{
  struct MHD_Connection *connection;	/* The MHD connection this message came from */
  gchar *method;
  gchar *url;
  gchar *contenttype;					/* Content-Type of the payload */

  json_t *response;					/* The response from the core */

  GHashTable *req_headers;   /* Loaded Headers */
  GHashTable *req_cookies;   /* Loaded Cookies */
  GHashTable *req_params;    /* URI parameters */

  GString *raw_body; /* Raw JSON body data */

  void* custom_data; /* Can be used by custom Handler for storing info */

  SkynetHTTP_RequestCompleteCallBack *complete_callback;

  char *user_agent;

  // JSON Specific
  json_t *j_request; /* Parsed JSON body data */
  gboolean signature_verified;
  char *api_key;
  char *msg_signature;
  char *client_id;
  char *api_version;
  char *header_nonce;
  char *session_id;
  char *user_id;
  char *remote_addr;
}skynet_http_msg;
typedef skynet_http_msg *skynet_http_msg_p;


// Callback function definition
typedef int
(*SkynetHTTP_RequestCallback) (void *cls,
struct MHD_Connection *connection,
const char *url,
const char *method,
const char *version,
const char *upload_data,
size_t *upload_data_size,
void **con_cls,
skynet_http_msg_p msg
);


// Nchan Callback
typedef int (*SkynetHTTP_Nchan_RequestCallback) (skynet_http_msg_p msg);



int skynet_http_return_forbidden_plain(skynet_http_msg *msg);
int skynet_http_return_success_x_accel(skynet_http_msg *msg, const char *redirUrl);
int skynet_http_return_success_x_accel_ex(skynet_http_msg *msg, const char *redirUrl,struct MHD_Response* response);
int skynet_http_return_success_plain(skynet_http_msg *msg);

/**
 *  Callback for Executing the request
 */
typedef int (*SkynetHTTP_Json_Exec) (skynet_http_msg_p msg);
/**
 *  Load API Key
 */
typedef char* (*SkynetHTTP_Api_Secret_Loader)(const char *apikey);


/**
* Callback for Nchan auth
* non-zero return value means auth error.
*/
typedef int (*SkynetHTTP_Nchan_Auth) (skynet_http_msg_p msg);
// Callback to return redirect for authenticated users
typedef int (*SkynetHTTP_Nchan_UserRedirectHeaders) (skynet_http_msg_p msg);
// Callback to find nchan channel for a user_id
typedef char* (*SkynetHTTP_Nchan_ChannelForUser) (const char *appId, const char* user_id);

typedef struct skynet_http_json_exec_s{
  SkynetHTTP_Json_Exec exec_func;
  SkynetHTTP_Api_Secret_Loader secret_loader_func;
}skynet_http_json_exec_s;
typedef skynet_http_json_exec_s* skynet_http_json_exec_p;


// Nchan executors loaded by group id
typedef struct skynet_http_nchan_exec_s{
  SkynetHTTP_Nchan_Auth nchan_auth;
  SkynetHTTP_Nchan_UserRedirectHeaders redir_headers;
  SkynetHTTP_Nchan_ChannelForUser channel_for_user;
}skynet_http_nchan_exec_s;
typedef skynet_http_nchan_exec_s* skynet_http_nchan_exec_p;

typedef int (*SkynetHTTP_Exec) (skynet_http_msg_p msg);

typedef struct skynet_http_exec_s{
  SkynetHTTP_Exec exec_func;
}skynet_http_exec_s;
typedef skynet_http_exec_s* skynet_http_exec_p;


/**
 * Structure for curl request response.
 * @var		typedef	struc
 * @global
 */
typedef struct _curlHttpResponse {
  GString *header;
  GString *body;
  char *errorMessage;
  int code;
  int httpCode;
}curlHttpResponse;
typedef curlHttpResponse* curlHttpResponsePtr;




/**
 * Internal function to write curl buffer to a Gstring structure.
 *
 * @author	mike
 * @since	v0.0.1
 * @version	v1.0.0	Friday, January 25th, 2019.
 * @global
 * @param	void  	*buffer
 * @param	size_t	size
 * @param	size_t	nmemb
 * @param	void  	*userp
 * @return	void
 */
size_t _curl_write_callback_gstring(void *buffer, size_t size, size_t nmemb, void *userp);

/**
 * Internal function to write curl header buffer to a Gstring*
 *
 * @author	mike
 * @since	v0.0.1
 * @version	v1.0.0	Friday, January 25th, 2019.
 * @global
 * @param	void  	*buffer
 * @param	size_t	size
 * @param	size_t	nmemb
 * @param	void  	*userp
 * @return	void
 */
size_t _curl_write_callback_header_gstring(void *buffer, size_t size, size_t nmemb, void *userp);



/**
 *  Dump a key value table to stdout.
 *
 *  @param table Table with key value contents.
 */
void skynet_http_msg_dump_key_val(GHashTable *table);

/**
 *  Verifies the message string if signed properly for the ID of signer.
 *
 *  @param message The message to verify.
 *  @param signature The signature to verify.
 *  @param hmac_id The Id for which we verifiy.
 *
 *  @return valid or not
 */
gboolean skynet_http_request_verify_signature(const char *message, const char *signature, const char *hmac_id);

/**
 *  Return a signature for given message and key.
 *
 *  @param message  message to sign
 *  @param hmac_id  id of signer
 *  @param hmac_key key for signing
 *
 *  @return the signature
 */
gchar* skynet_http_message_sign(const char *message, const char *hmac_id, const char *hmac_key);


/**
 *  Helper to return success to client.
 *
 *  @param msg     Message object.
 *  @param method   Name of the response.
 *  @param code Code of response.
 *  @param message Message that goes with if code if any.

 *  @return Status
 */
int skynet_http_return_success(skynet_http_msg *msg, const char *method,int code, const char *message);



/**
 *  Helper to return error to client.
 *
 *  @param msg MSG Object.
 *  @param message The sender copies this buffer.
 *  @param code HTTP Code.
 *
 *  @return
 */
int skynet_http_return_error(skynet_http_msg *msg, char *message,int code);

/**
 *  Cleanup after HTTP request is done.
 *
 */
void skynet_http_request_completed(void *cls, struct MHD_Connection *connection, void **con_cls, enum MHD_RequestTerminationCode toe);

/**
 *  Callback which is called for each connection.
 *
 *  @param cls
 *  @param connection
 *  @param url
 *  @param method
 *  @param version
 *  @param upload_data
 *  @param upload_data_size
 *  @param con_cls
 *
 *  @return
 */
int skynet_http_answer_connection(void *cls, struct MHD_Connection *connection,
                                  const char *url,
                                  const char *method, const char *version,
                                  const char *upload_data,
                                  size_t *upload_data_size, void **con_cls);


/**
 *  Create a HTTP Server - This is the server function.
 *
 *  @param ip_address IP Address to Bind
 *  @param port       port to listen to
 *  @param path       path called
 *  @param accept_cb  Accept policy can be NULL.
 *  @param request_callback     Callback called for each request.
 *
 *  @return Server struct ptr
 */
struct MHD_Daemon *skynet_http_create_server(const char *ip_address, int port, const char *path,MHD_AcceptPolicyCallback *accept_cb,SkynetHTTP_RequestCallback *request_callback);


/**
 *  Return a JSON to client.
 *
 *  @param result
 *  @param msg
 *  @param http_code
 */
void skynet_http_return_json(json_t *result,skynet_http_msg_p msg,int http_code);

/**
 *  Return a JSON Server error response.
 *
 *  @param msg           Message object
 *  @param json_err_msg  The description of the Error
 *  @param json_err_code The code of the Error
 *  @param http_code     HTTP Code to return.
 */
void skynet_http_json_server_return_err(skynet_http_msg_p msg,const char *json_err_msg,int json_err_code,int http_code);


/**
 *  Return the value for a header.
 *
 *
 *  @param msg         Message object.
 *  @param header_name Name of the Header to return value for.
 *
 *  @return returns a string or NULL if not found.
 */
char* skynet_http_get_header_value(skynet_http_msg_p msg,const char *header_name);


/**
 *  Create a new empty Json response
 *
 *  @param method   Name of response (rpc caller name or something like that)
 *  @param success  Is a success response or failure response.
 *  @param code     If failed code of failure.
 *  @param message  If needed add this messag to the object.
 *
 *  @return The new json object.
 */
json_t* skynet_http_json_response_new(const char *method,gboolean success,int code,const char *message);

/**
 *  Append payload field to existing response object.
 *
 *  @param response response object.
 *  @param payload  Object to be appended. The append functions steals the reference so no need to call decref.
 */
void skynet_http_json_response_append_payload(json_t **response,json_t *payload);


/**
 *  Make a request to Skynet HTTP Server.
 *
 *  @param uri       Address to make request to.
 *  @param j_request Json data to send with request.
 *  @param response  Response of the request.
 *
 *  @return Return success of request.
 */
gboolean skynet_http_json_request(const char *uri,json_t *j_request,skynet_http_response_p *response);


/**
 *  Create a new Skynet HTTP JSON Server.
 *
 *  @param port                      Port to listen to
 *  @param default_context_executors Default hash table with functions for executing default context.
 *
 *  @return Success of server creation.
 */
gboolean skynet_http_json_server_new(int port,GHashTable *default_context_executors,gpointer global_manager);

/**
 *  Add a new executor to JSON HTTP Server.
 *
 *  @param context           The context for which we add the function. Context is main URL e.g. /bank or /transcoder
 *  @param executor_function Ptr to function to exec.
 *
 *  @return Success of add.
 */
gboolean skynet_http_json_server_add_executor(const char *context,skynet_http_json_exec_p executor);

/**
 *  Executor by context.
 *
 *  @param context
 *
 *  @return
 */
skynet_http_json_exec_p skynet_http_json_server_fetch_executor(const char *context);



/**
 *  Add a new executor to Nchan HTTP Server.
 *
 *  @param context           The context for which we add the function. Context is the Nchan group.
 *  @param executor_function Ptr to struct holding the functions to exec.
 *
 *  @return Success of add.
 */
gboolean skynet_http_nchan_server_add_executor(const char *context,skynet_http_nchan_exec_p executor);

/**
 *  Executor by context.
 *
 *  @param context
 *
 *  @return skynet_http_nchan_exec_p
 */
skynet_http_nchan_exec_p skynet_http_nchan_server_fetch_executor(const char *context);

/**
 * Create a new Nchan HTTP Listener.
 * The NChan listener authorizes users to subscribe to a specific channel.
 * The ChannelId is generated with a xxHash from the user's id.
 *
 * @param port Port to listen to
 * @param default_content_executors Holds the executor functions for requests.
 * @return Success of server creation.
 */
gboolean skynet_http_nchan_server_new(int port,GHashTable *default_context_executors,gpointer global_manager);


/**
 * Send a message to user, this function finds the apropriate channel id for the given user
 * with the executors helpers.
 *
 * @param appId Application Id
 * @param user_id The userId to send the message to.
 * @param message The message to send.
 */
gboolean skynet_http_nchan_push_message_user(const char *appId,const char *user_id,const char *message);


/**
 * Send a message to user, this function finds the apropriate channel id for the given user
 * with the executors helpers.
 *
 * @param appId Application Id
 * @param channel_id ID of Channel to send message to.
 * @param message The message to send.
 */
gboolean skynet_http_nchan_push_message_channel_id(const char *appId,const char *channel_id,const char *message);

/**
 * Destroy a given channel.
 * @param appId Application Id.
 * @param channel_id Id of channel to destroy.
 *
 * */
void skynet_http_nchan_channel_destroy(const char *appId, const char *channel_id);



/**
 * Send a message to a specific session, the channel is the sessionId.
 * Usable for guest sessions.
 *
 * @param appId Application Id
 * @param session_id The userId to send the message to.
 * @param message The message to send.
 */
gboolean skynet_http_nchan_push_message_session(const char *appId,const char *session_id,const char *message);



/**
* Fetch channel info in json format by userid.
*
* @return int If non 0 then the call failed.
*/
int skynet_http_nchan_channel_info_userid(const char *appId, const char *userId,json_t **j_channel_info);




// gboolean skynet_http_client_get(const char *requestUrl,GHashTable *requestHeaders, GString **resultBuf, GError **err);



/**
 * skynet_http_client_delete.
 *
 * @author	Unknown
 * @since	v0.0.1
 * @version	v1.0.0	Friday, January 25th, 2019.
 * @global
 * @param	char      	*requestUrl
 * @param	ghashtable	*requestHeaders
 * @return	curlHttpResponsePtr
 */
curlHttpResponsePtr skynet_http_client_delete(const char *requestUrl,GHashTable *requestHeaders);

/**
 * skynet_http_client_get.
 *
 * @author	Unknown
 * @since	v0.0.1
 * @version	v1.0.0	Friday, January 25th, 2019.
 * @global
 * @param	char      	*requestUrl
 * @param	ghashtable	*requestHeaders
 * @return	curlHttpResponsePtr
 */
curlHttpResponsePtr skynet_http_client_get(const char *requestUrl,GHashTable *requestHeaders);


/**
 * skynet_http_client_get_curlheaders.
 *
 * @author	Unknown
 * @since	v0.0.1
 * @version	v1.0.0	Friday, January 25th, 2019.
 * @global
 * @param	char  	*requestUrl
 * @param	struct	curl_slis
 * @return	curlHttpResponsePtr
 */
curlHttpResponsePtr skynet_http_client_get_curlheaders(const char *requestUrl,struct curl_slist headersList);

/**
 * skynet_http_curl_response_init.
 *
 * @author	Unknown
 * @since	v0.0.1
 * @version	v1.0.0	Friday, January 25th, 2019.
 * @global
 * @return	void
 */
curlHttpResponsePtr skynet_http_curl_response_init();
/**
 * @var		mixed	voi
 * @return void
 * @global
 */
void                skynet_http_curl_response_free(curlHttpResponsePtr response);

/**
 * Init of Nchan Publisher
 *
 * @author	Unknown
 * @since	v0.0.1
 * @version	v1.0.0	Friday, January 25th, 2019.
 * @global
 * @return	void
 */
void skynet_http_nchan_init_publisher();



END_EXTERN_C()

#endif /* SkynetHTTP_hpp */
