/*
 * utils.h
 *
 * Author: Miha Nedok <mike@mike.si>
 */

#ifndef SKYNETLIB_UTILS_H_
#define SKYNETLIB_UTILS_H_

#include <glib.h>


/****** UTILS *********/

#include <uuid/uuid.h>

// UUID type for non Mac platform
#ifndef _DARWIN_C_SOURCE
typedef char uuid_string_t[37];
#endif



char* mp_strdup(const char *str);


unsigned long long skynet_xxhash(const void *buffer,size_t length);
gchar* skynet_hash_hex(const void *buffer,size_t length);



/**
 * Create new UUID and return as string
 */
gchar* genuuid();

gchar* genrandom(int len);



/**
 * Parse given string of headers into destArray, the array contains
 * mpHttpHeaders members, if parsing successful the free is automaticaly given
 */
void parseHttpHeaders(const char* rawHeaders,GPtrArray** destArray);
void _freeMpHttpHeader(void* m);
void dumpHeaders(GPtrArray* headers);
gboolean parseRawCookies(const char *src,GHashTable **cookieJar);

time_t nowtime();




/* Conversions */
typedef unsigned long uint32;
void uint32_pack(char *s,uint32 u);
void uint32_pack_c(char *s,uint32 u);
void uint32_unpack(char *s,uint32 *u);
void uint32_unpack_c(char *s,uint32 *u);
/* End Conversions */

/********* END UTILS ***/

/*** Time and Timing ***/
#define TIME_MICRO_IN_SEC 1000000.00

double mpTimerStart();
double mpTimerStop(double start_time_val);

/*** End Time and Timing ***/


/** Hash Tables **/

#define G_STR_HASH_TABLE_NEW g_hash_table_new(g_str_hash,g_str_equal)
#define G_INT_HASH_TABLE_NEW g_hash_table_new(g_direct_hash,g_direct_equal)

/** End Hash Tables **/


char *hmacSignMessage(const char *message,size_t len, const char *secret);
gboolean hmacVerifySignature(const char *message, size_t len,const char *secret,const char *signatureHex);
gboolean mp_encrypt_buffer(char *buf,int len,char *key,char **result, int *resultLen,gboolean useBase64,gboolean convertKey);

/**
 * privateKey and publicKey should be freed with g_free
 */
void SkynetGenerateRSAKeypair(const char *pass,char **privateKey, char **publicKey);
void SkynetGenerateX509Cert(const char *pass,const char *userid,const char *email, const char *organization);
void SkynetX509EncryptMessage(json_t *payload);


/*** Compression ***/

gboolean mpReadCompressedFile(const char* fileName,GString **uncompressedBuf);
gboolean mpWriteCompressedFile(const char *fileName,GString* buf);

/*** End Compression ***/

#define genguid genuuid


guint32 mpRand();


/************************** JSON Macros ***************************/

#define J_STRVAL(strobject) json_string_value((json_t*)strobject)
#define J_OBJ_G(json_object,key) json_object_get((json_t*)json_object,key)
#define J_INTVAL(intobject) json_integer_value((json_t*)intobject)

#define J_OBJ_GET_STR(json,key) json_string_value(json_object_get((json_t*)json,key))
#define J_OBJ_GET_STR_COPY(json,key) g_strdup((json_is_string(json_object_get((json_t*)json,key)))?J_OBJ_GET_STR((json_t*)json,key):NULL)
#define J_STRCOPY(jObj) mp_strdup((char*)json_string_value((json_t*)jObj))
#define J_OBJ_GET_INT(json,key) json_integer_value(json_object_get((json_t*)json,key))
#define J_OBJ_GET_BOOL(json,key) json_boolean(json_object_get((json_t*)json,key))


/**
 * Add new to json object and free source char ptr
 * if str ptr is null then add null to object
 */
#define J_OBJ_ADD_NEW_FREE(j,key,str){\
	json_object_set_new(j,key,(str!=NULL)?json_string(str):json_null());\
	if(str){g_free(str);}\
}

#define J_OBJ_ADD_NEW(j,key,str){\
	json_object_set_new(j,key,(str!=NULL)?json_string(str):json_null());\
}

#define J_OBJ_ADD_NEW_INT(j,key,i){\
	json_object_set_new(j,key,json_integer(i));\
}


#define J_OBJ_GET_BEGIN(jObj){\
	json_t *jjjtmpOBJ = (json_t*)jObj;
#define J_OBJ_G_E(key) json_object_get(jjjtmpOBJ,key)
#define J_OBJ_GET_END() }



#define J_DECREF(json){\
	if(json!=NULL)\
		json_decref(json);\
}


/*************************** END JSON Macros ***************************/


/*************** CONFIG STUFF ****************************************/
/**
 * 	phpRelayUrl =  J_OBJ_G(configRoot,"phpRelayUrl");
	if (phpRelayUrl != NULL){
		config->phpRelayUrl = g_strdup(J_STRVAL(phpRelayUrl));
	}
 *
 */
#define CONFIG_SETTING_COPY_STR_EX(destJson,jsonRoot,jsonCfgKey,destStr){\
 destJson = J_OBJ_G(jsonRoot,jsonCfgKey);\
 if (destJson != NULL){\
	 destStr = g_strdup(J_STRVAL(destJson));\
 }\
}


#define CONFIG_SETTING_COPY_STR(jsonRoot,jsonCfgKey,destStr){\
 json_t *tmpJson = J_OBJ_G(jsonRoot,jsonCfgKey);\
 destStr = NULL;\
 if (tmpJson != NULL){\
	 destStr = g_strdup(J_STRVAL(tmpJson));\
 }\
}



/*** Mongo Stuff  ***/

#define BEGIN_BSON(bsonObj){\
	bson_init(bsonObj);\

#define END_BSON }


#define MONGO_WRITE_CONCERN(writeConcern){\
		 mongo_write_concern *wcObj =  write_concern;\
     mongo_write_concern_init( wcObj );\
     wcObj->w = 1;\
     mongo_write_concern_finish( wcObj );\
}


#define BSON_BEGIN_B(b){\
	bson* bsonObj = b;
#define BSON_END_B }



#define BSON_APPEND_S(key,val) bson_append_string(bsonObj,key,val)
#define BSON_APPEND_SL(key,val,len) bson_append_string_n(bsonObj,key,val,len)
#define BSON_APPEND_NEWOID(key)	bson_append_new_oid(bsonObj,key)
#define BSON_APPEND_I(key,val) bson_append_int(bsonObj,key,val)


#define BSON_OBJ_DEST(dbname,objectName) size_t dbLen = strlen(dbname),objLen = strlen(objectName);\
	char mongoCollectObjName[objLen+dbLen+2];\
	memcpy(mongoCollectObjName,dbname,dbLen);\
	memcpy(mongoCollectObjName+dbLen,'.',1);\
	memcpy(mongoCollectObjName+dbLen+1,objectName,objLen);\
	mongoCollectObjName[dbLen+objLen+1] = '\0';\



#define BSON_APPEND_SAFE(destObj,srcObj,fieldName){\
if (srcObj.hasElement(fieldName)){\
	destObj.append(srcObj.getField(fieldName));\
} else {\
	destObj.appendNull(fieldName);\
}\
}

#define BSON_APPEND_SAFE_STR(destObj,srcObj,fieldName,defValue){\
if (srcObj.hasElement(fieldName)){\
destObj.append(srcObj.getField(fieldName));\
} else {\
destObj.append(fieldName,defValue);\
}\
}

#define BSON_APPEND_SAFE_ARR(destObj,srcObj,fieldName){\
if (srcObj.hasElement(fieldName)){\
destObj.append(srcObj.getField(fieldName));\
} else {\
BSONArrayBuilder aTmp;\
destObj.appendArray(fieldName,aTmp.arr());\
}\
}



/*** End Mongo Stuff ***/



#endif /* UTILS_H_ */
