/*
 * SkynetTypes.h
 *
 * Author: Miha Nedok <mike@mike.si>
 */




#ifndef SKYNETTYPES_H_
#define SKYNETTYPES_H_


#define SK_PARAM_STRING JSON_STRING
#define SK_PARAM_INT    JSON_INTEGER
#define SK_PARAM_OBJECT JSON_OBJECT
#define SK_PARAM_ARRAY  JSON_ARRAY


#define SK_PARAM_REQUIRED 1
#define SK_PARAM_NONEMPTY 2
#define SK_PARAM_POSITIVE 3


#define SKYNET_ERROR_TEXT_LEN 1024

typedef struct skynet_error_t{
  char msg[SKYNET_ERROR_TEXT_LEN];
  int code;
}skynet_error_t;


struct skynet_serverfunc_parameter {
  const char* name;
  int ptype;
  unsigned int flags;
};

typedef struct _relayRequestResponse {
  int httpCode;
  GString* rawResponse;
  GString* headerRawData;
  GPtrArray* responseHeaders;
  gboolean success;
}RelayRequestResponse;


typedef struct _skynetAppSetting{
  char *appId;
  char *appName;
  char *dbName;
}skynetAppSetting;

/**
 * Service structure - holds
 * all the configuration variables needed
 */
typedef struct _ServiceConfig {
  gboolean master;
  char *masterIp;
  char *nodeName;
  char *serviceName; // Friendly Service Name
  char *endpointPath; // The distinction path for the service to identify
  char *phpRelayUrl;
  char *ldapServer; // LDAP or Windows AD Server
  char *sessionName; // The cookie name for session
  char *mongoHost;
  char *mongoDb;
  gboolean mongoDisabled;
  char *mysqlHost;
  char *mysqlUser;
  char *mysqlPass;
  char *mysqlDb;
  int websocketPort;
  int phpServerPort;
  int jsonServerPort;
  int internalJsonServerPort;
  gboolean messagingEnbaled;
  char *smtpServer;
  char *smtpUsername;
  char *smtpPassword;
  char *smsInterface;
  char *actionsDictFilePath;
  GHashTable *actionsTable;
  GPtrArray *modules;
  char *phpDocumentRoot;
  char *adminHmacSecret;
  char *functionDumpFile;
  char *appsMapFile;
  char *vpnJsonServerIpAddress;
  int vpnJsonServerPort;
  char *vpnJsonInernalServerIpAddress;
  int vpnJsonInternalServerPort;
}ServiceConfig;

/**
 * Skynet modules
 */

#define SKYNET_MODULE_WS_RELAY_FUNC "m_skynet_ws_message_relay_func"
#define SKYNET_MODULE_SERVER_RELAY_FUNC "m_skynet_server_relay_func"
#define SKYNET_MODULE_VERSION_FUNC "m_skynet_version_func"
#define SKYNET_MODULE_INIT_FUNC "m_skynet_init"



typedef struct {
  gboolean success;
}SkynetModuleResult;


extern void printVersion();

/**
 * Skynet Module support
 * The module is aware of the user and session.
 */
typedef struct {
  short messageType;
  json_t *jsonPayload;
  gboolean success;
  json_t *jsonResult;
  char *sessionId;
  char *userId;
  char *messageName;
}SkynetModuleMessage;


#define SKYNET_SERVER_MESSAGE_TYPE_JSON	1
#define SKYNET_SERVER_MESSAGE_TYPE_BSON	2
/**
 * Skynet Server Message
 * is a structure to handle requests from
 * PHP or Web clients in a uniform way
 *
 * Members shall not be freed, except:
 * jsonResult and bsonResult of not NULL
 */
typedef struct _skynetServerMessage {
  short source;	// Source system - PHP, WS or other
  char *messageName; // Name of the message from client
  void *payload; // Input payload in json or BSON structure
  char* origPayload; // Original payload
  gboolean success; // Indicates execution success
  double execTime; // The execution time for this message - self timed
  char *sessionId; // SessionId of client requesting
  char *userId; // UserId of client requesting
  char *remoteAddr; // The address of client requesting
  json_t *jsonResult; // The executor writes result to this object
  GString *stringResult; // The executor writes here a string result
  ServiceConfig *config; // Ptr to config
  void* bsonResult; // Ptr to BSONObj
  void* paramsBucket; // Ptr to BSONObjBuilder
  void* sessManager; // Ptr to Server's SessionManager
  char *appId;
  struct SkynetServerCommand *ptrToCmd; // Pointer to function to execute
  int messageType;
}SkynetServerMessage;

#define ss_param_type_string            1;
#define ss_param_type_int		            2;
#define ss_param_type_array             3;
#define ss_param_type_object            4;
#define ss_param_type_bson	            5;


#define SKYNET_ALLOWED_SOURCE_WS        1
#define SKYNET_ALLOWED_SOURCE_SRV       2
#define SKYNET_ALLOWED_SOURCE_BOTH      3
#define SKYNET_ALLOWED_SOURCE_INTERNAL  4
#define SKYNET_ALLOWED_SOURCE_ALL       5
typedef void SkynetServerCommandProc(SkynetServerMessage*);
struct SkynetServerCommand {
  const char *name;
  SkynetServerCommandProc *cmd;
  short sourceAllowed;
  gboolean mustBeAuthed;
  char *paramsList;
  GHashTable *parameters;
  short source;
};


SkynetServerMessage* skynet_server_message_alloc();

//* Function ptrs */

typedef void  (*SkynetWsMessageRelay) (SkynetModuleMessage **msg, SkynetModuleResult **result);
typedef void  (*SkynetServerMessageRelay) (SkynetModuleMessage **msg, SkynetModuleResult **result);
typedef void  (*SkynetModuleExec) (SkynetServerMessage *msg);
typedef struct SkynetServerCommand*  (*SkynetModuleFetchFunction) (const char *commandName);
typedef char* (*SkynetModuleVersion) (void);
typedef void  (*SkynetModuleInit) (void *ptrSessionManager);

/**
 * End Skynet Modules
 */


typedef struct _mpHttpHeader{
  char *name;
  char *value;
}mpHttpHeader;



#endif /* SKYNETTYPES_H_ */

#ifndef TYPES_H
#define	TYPES_H

typedef struct _skynet_error_obj_s {
  char* logMsg;
  int   type;
  char* displayMsg;
}skynet_error_obj_s;

#define SKYNET_ERROR(errObj,logMsg,displayMsg,type){\
errObj.logMsg = logMsg;\
errObj.type		= type;\
errObj.displayMsg = displayMsg;\
}


typedef struct _skynet_global_stats {
  long commandsProcessed;
  long commandsFailed;
  long slowCommands;
  time_t startTime;
}skynet_global_stats;



typedef struct _log_audit_s {
  time_t remoteTime;				// Time of Event by remote peer
  time_t createTime;				// Our time
  time_t verificationTime;	// Verification Time of Event
  char *sessionId;					// If from session
  char *userId;							// If having userId
  char *remoteAddr;					// Remote IP of Audit
  char *context;						// Context of audit (e.g. user actions)
  char *type;								// Type of Audit
  char *sourceSystem;				// Source system id
  char *targetSystem;				// If having target system
  char *originator;					// Originator of Event
  char *event;							// Event ID
  char *message;						//
  char *additionalMessage;	// Serialized additional data
}log_audit_s;



/**
 * Session Entry - Single user session structure
 */
typedef struct _SessionEntry {
  gchar *sessionId;
  gchar *userId;
  gboolean isMobile;
  char *deviceId; // Mobile clients provide hash id of device
  gboolean isAdmin;
  time_t createdTime;
  time_t lastActivity;
  int expireTime;
  int permanent;
  gchar *remoteAddr; // last address when called
  gchar *companyId; // If in database - set when loging in - if need new then load company
  gchar *companyType; // If in database - set when loging in - if need new then load company
  GString *data; // PHP Data
  json_t *jSessionData; // Json Object holding session data accessible to Skynet
  GHashTable *bsonBucket;
  gchar *strUserType; // If in database - set when loging in - if need new then load user
  gchar *applicationId; // App created the session and is valid for
  // Time Authorization is per session based for each user
  gboolean isTimeAuthorized;
  time_t timeAuthorizedTime;
}SessionEntry;
SessionEntry* initSessionEntry();

typedef struct _skynet_websocket_action_s{
  char *name;
  int id;
  char *route;
}skynet_websocket_action_s;

#define SKYNET_WEBSOCKET_ACTION_S_INIT(ptr){\
ptr = (skynet_websocket_action_s*)g_new0(skynet_websocket_action_s,1);\
ptr->name = NULL;\
ptr->id = 0;\
ptr->route = NULL;\
}


typedef struct {
  size_t all;
  size_t loggedon;
  size_t anon;
}online_info_s;


typedef enum SkynetWebsocketActions {
  buddyList = 1000,
  
  userData = 1001,
  newMsg = 2000,
  newConversation = 2001,
  getConversations = 2002
}SkynetWebsocketActions;




/* Global Register Types */
#define SKYNET_REGISTER_TYPE_RECORDER (10000 * 1000)
#define SKYNET_REGISTER_TYPE_TRANSCODER (1000 * 1001)
#define SKYNET_REGISTER_TYPE_MEDIA_GW (1000 * 1002)
#define SKYNET_REGISTER_TYPE_RESTREAMER (1000 * 1003)
#define SKYNET_REGISTER_TYPE_APICLIENTS (2000 * 1000)



#ifdef __cplusplus
//#include "SessionManager.h"
#endif

typedef struct _phpServerData {
#ifdef __cplusplus
  //	SessionManager* sessionManager; // global session manager
#endif
  ServiceConfig *config;
}phpServerData;



#endif	/* TYPES_H */
